"""

circonus.client
~~~~~~~~~~~~~~~

"""

from datetime import datetime
from functools import wraps
from posixpath import sep as pathsep
from urlparse import SplitResult, urlunsplit

import logging
import json

from circonus.annotation import Annotation
from circonus.collectd.cpu import get_cpu_graph_data
from circonus.collectd.memory import get_memory_graph_data
from circonus.collectd.interface import get_interface_graph_data
from circonus.tag import get_tags_with, get_telemetry_tag, is_taggable
from requests.exceptions import HTTPError

import requests


API_PROTOCOL = "https"
API_LOCATION = "api.circonus.com"
API_VERSION = 2
API_BASE_SPLIT = SplitResult(scheme=API_PROTOCOL, netloc=API_LOCATION, path="/v%d" % API_VERSION, query="", fragment="")
API_BASE_URL = urlunsplit(API_BASE_SPLIT)

log = logging.getLogger(__name__)


def get_api_url(resource_type_or_cid):
    """Get a valid fully qualified Circonus API URL for the given resource type or ``cid``.

    :param str resource_type_or_cid: The resource type or ``cid`` representing a specific resource.
    :return: The API URL.
    :rtype: :py:class:`str`

    """
    return pathsep.join([API_BASE_URL, resource_type_or_cid.strip(pathsep)])


def with_common_tags(f):
    """Decorator to ensure that common tags exist on resources.

    Only resources which support tagging via the Circonus API will be affected.

    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        client, cid, data = args[:3]
        common_tags = list(client.common_tags)
        if is_taggable(cid):
            if "type" in data:
                common_tags.append(get_telemetry_tag(data))

            if data.get("tags"):
                updated_tags = get_tags_with(data, common_tags)
                if updated_tags:
                    data.update({"tags": updated_tags})
            else:
                data["tags"] = common_tags
        return f(*args, **kwargs)
    return wrapper


def log_http_error(f):
    """Decorator for :py:mod:`logging` any :class:`~requests.exceptions.HTTPError` raised by a request."""
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            r = f(*args, **kwargs)
            r.raise_for_status()
        except HTTPError as e:
            log.error("%s: %s (%s)", e.response.status_code, e.response.json()["code"],
                      e.response.json()["message"])
        return r
    return wrapper


class CirconusClient(object):
    """Construct a :class:`CirconusClient`.

    :param str api_app_name: The Circonus API application name.
    :param str api_token: The Circonus API token.
    :param list common_tags: (optional) The :py:class:`str` tags to apply to all resources.
    :rtype: :class:`CirconusClient`

    Usage::

        >>> from circonus import CirconusClient
        >>> circonus = CirconusClient("my-circonus-app", "generated-by-circonus-ui")
        >>> response = circonus.get("/user/current")
        >>> response.json()
        {u'_cid': u'/user/1234',
         u'contact_info': {u'sms': u'', u'xmpp': u''},
         u'email': u'user@example.com',
         u'firstname': u'Ewe',
         u'lastname': u'Sure'}

    """

    def __init__(self, api_app_name, api_token, common_tags=None):
        self.api_app_name = api_app_name
        self.api_token = api_token
        self.api_headers = {
            "Accept": "application/json",
            "X-Circonus-App-Name": self.api_app_name,
            "X-Circonus-Auth-Token": self.api_token
        }
        if common_tags is None:
            self.common_tags = []
        else:
            self.common_tags = common_tags

    @log_http_error
    def get(self, resource_type_or_cid, params=None):
        """Get the resource at resource type or ``cid`` via :func:`requests.get`.

        :param str resource_type_or_cid: The resource type or ``cid`` representing a specific resource.
        :param dict params: (optional) The parameters to pass to :func:`requests.get`.
        :rtype: :class:`requests.Response`

        If a resource type is given, e.g., ``/user``, a :py:class:`list` of resources will exist in the :class:`~requests.Response` JSON.

        If a ``cid`` is given, e.g., ``/check_bundle/123456``, a single resource will exist in the :class:`~requests.Response` JSON.

        """
        return requests.get(get_api_url(resource_type_or_cid), params=params, headers=self.api_headers)

    @log_http_error
    def delete(self, cid, params=None):
        """Delete the resource at ``cid`` via :func:`requests.delete`.

        :param str cid: The resource to delete.
        :param dict params: (optional) The parameters to pass to :func:`requests.delete`.
        :rtype: :class:`requests.Response`

        """
        return requests.delete(get_api_url(cid), params=params, headers=self.api_headers)

    @with_common_tags
    @log_http_error
    def update(self, cid, data):
        """Update the resource at ``cid`` with ``data`` via :func:`requests.put`.

        :param str cid: The resource to update.
        :param dict data: The data used to update the resource.
        :rtype: :class:`requests.Response`

        """
        return requests.put(get_api_url(cid), data=json.dumps(data), headers=self.api_headers)

    @with_common_tags
    @log_http_error
    def create(self, resource_type, data):
        """Create the resource type with ``data`` via :func:`requests.post`.

        :param str resource_type: The resource type to create.
        :param dict data: The data used to create the resource.
        :rtype: :class:`requests.Response`

        """
        return requests.post(get_api_url(resource_type), data=json.dumps(data), headers=self.api_headers)

    def update_with_tags(self, cid, new_tags):
        """Update the resource at ``cid`` to have ``new_tags`` added to it via :func:`update`.

        :param str cid: The resource to update.
        :param list new_tags: The :py:class:`str` tags to add to the resource.
        :rtype: :class:`requests.Response` or :py:const:`False`

        :py:const:`False` is returned in instances where a request to update the resource was not necessary because
        the resource is already tagged with ``new_tags``.

        """
        r = False
        if is_taggable(cid):
            r = self.get(cid)
            tags = get_tags_with(r.json(), new_tags)
            if tags:
                r = self.update(cid, {"tags": tags})
        return r

    def annotation(self, title, category, description="", rel_metrics=None):
        """Context manager and decorator for creating :class:`~circonus.annotation.Annotation` instances.

        :param str title: The title.
        :param str category: The category.
        :param str description: (optional) The description.
        :param list rel_metrics: (optional) The :py:class:`str` names of metrics related to this annotation.
        :rtype: :class:`~circonus.annotation.Annotation`

        If ``rel_metrics`` is given, the metric names should be specified in the fully qualified format
        ``<digits>_<string>`` as required by the Circonus API documentation for `annotation
        <https://login.circonus.com/resources/api/calls/annotation>`_.

        """
        return Annotation(self, title, category, description, rel_metrics)

    def create_annotation(self, title, category, start=None, stop=None, description="", rel_metrics=None):
        """Create an :class:`~circonus.annotation.Annotation` instance immediately.

        :param str title: The title.
        :param str category: The category.
        :param datetime.datetime start: (optional) The start time.
        :param datetime.datetime stop: (optional) The stop time.
        :param str description: (optional) The description.
        :param list rel_metrics: (optional) The :py:class:`str` names of metrics related to this annotation.
        :rtype: :class:`~circonus.annotation.Annotation`

        *Note*: sub-second annotation time resolution is not supported by Circonus.

        ``stop`` defaults to the value of ``start`` if it is not given.

        If ``rel_metrics`` is given, the metric names should be specified in the fully qualified format
        ``<digits>_<string>`` as required by the Circonus API documentation for `annotation
        <https://login.circonus.com/resources/api/calls/annotation>`_.

        """
        a = Annotation(self, title, category, description, rel_metrics)
        a.start = datetime.utcnow() if start is None else start
        a.stop = a.start if stop is None else stop
        a.create()
        return a

    def create_collectd_cpu_graph(self, target):
        """Create a CPU graph from a ``collectd`` check bundle for ``target``.

        :param str target: The target of the check bundle to filter for.
        :rtype: :class:`requests.Response`

        ``target`` is used to filter ``collectd`` check bundles.

        """
        r = self.get("check_bundle", {"f_target": target, "f_type": "collectd"})
        r.raise_for_status()
        check_bundle = r.json()[0]
        data = get_cpu_graph_data(check_bundle)
        return self.create("graph", data)

    def create_collectd_memory_graph(self, target):
        """Create a memory graph from a ``collectd`` check bundle for ``target``.

        :param str target: The target of the check bundle to filter for.
        :rtype: :class:`requests.Response` or :py:const:`None`

        ``target`` is used to filter ``collectd`` check bundles.

        :py:const:`None` is returned if no data to create the graph could be found for ``target``.

        """
        r = self.get("check_bundle", {"f_target": target, "f_type": "collectd"})
        r.raise_for_status()
        check_bundle = r.json()[0]
        data = get_memory_graph_data(check_bundle)
        return self.create("graph", data) if data else None

    def create_collectd_interface_graph(self, target):
        """Create an interface graph from a ``collectd`` check bundle for ``target``.

        :param str target: The target of the check bundle to filter for.
        :rtype: :class:`requests.Response` or :py:const:`None`

        ``target`` is used to filter ``collectd`` check bundles.

        :py:const:`None` is returned if no data to create the graph could be found for ``target``.

        """
        r = self.get("check_bundle", {"f_target": target, "f_type": "collectd"})
        r.raise_for_status()
        check_bundle = r.json()[0]
        data = get_interface_graph_data(check_bundle)
        return self.create("graph", data) if data else None



"""

circonus.graph
~~~~~~~~~~~~~~

Create graph data from a check bundle.

"""


from circonus.tag import get_tags_with, get_telemetry_tag


def get_graph_data(check_bundle, datapoints, custom_data=None):
    """Get graph data for ``check_bundle`` and  ``datapoints``.

    :param dict check_bundle: The check bundle to create graph data for.
    :param list datapoints: The datapoints to include in the graph data.
    :param dict custom_data: (optional) The custom data to include in the graph data.
    :rtype: :py:class:`dict`

    Merge common graph data with ``custom_data``.  Add the telemetry tag based on the ``type`` attribute of
    ``check_bundle``.

    """
    if custom_data is None:
        custom_data = {"tags": []}
    elif "tags" not in custom_data:
        custom_data["tags"] = []

    tags = get_tags_with(custom_data, [get_telemetry_tag(check_bundle)])
    common = {"datapoints": datapoints}
    if tags:
        common["tags"] = tags

    custom_data.update(common)
    return custom_data



"""

circonus.collectd.cpu
~~~~~~~~~~~~~~~~~~~~~

Create graph data from a ``collectd`` check bundle.

"""

from collections import OrderedDict
from copy import deepcopy
from itertools import chain

import re

from circonus.graph import get_graph_data
from circonus.metric import get_datapoints, get_metrics, get_metrics_sorted_by_suffix
from circonus.util import get_check_id_from_cid


CPU_METRIC_SUFFIXES = ["steal", "interrupt", "softirq", "system", "wait", "user", "nice", "idle"]
"""Ordered list of metric suffixes.

This list is used to filter and sort metrics in preparation for creating a graph.

"""

CPU_METRIC_RE = re.compile(r"""
^cpu                            # Starts with "cpu"
`.*`                            # Anything in between
""", re.X)
"""A compiled regular expression which matches ``collectd`` metrics."""

CPU_NUMBER_RE = re.compile(r"""
^cpu                            # Starts with "cpu"
`                               # Delimiter
(?P<number>\d+)                 # Number
`                               # Delimiter
""", re.X)
"""A compiled regular expression which captures CPU number from the CPU metric."""


def _get_cpus(metrics):
    """Get a list of strings representing the CPUs available in ``metrics``.

    :param list metrics: The metrics used to look for CPUs.
    :rtype: :py:class:`list`

    The returned strings will begin with the CPU metric name. The list is sorted in ascending order.

    """
    cpus = list({m["name"].rpartition("cpu")[0] for m in metrics})
    cpus.sort()
    return cpus


def get_cpu_metrics(metrics):
    """Get a sorted list of metrics from ``metrics``.

    :param list metrics: The metrics to sort.

    The CPU metrics are sorted by:

    #. Name, ascending
    #. Explicit suffix, i.e., :const:`~circonus.collectd.cpu.CPU_METRIC_SUFFIXES`

    """
    cpus = _get_cpus(metrics)
    cpu_metrics = OrderedDict.fromkeys(cpus)
    for cpu in cpus:
        cpu_metrics[cpu] = get_metrics_sorted_by_suffix((m for m in metrics if m["name"].startswith(cpu)),
                                                        CPU_METRIC_SUFFIXES)
    return list(chain.from_iterable(cpu_metrics.values()))


def get_stacked_cpu_metrics(metrics, hide_idle=True):
    """Get metrics with the ``stack`` attribute added.

    :param list metrics: The metrics to stack.
    :param bool hide_idle: (optional) Hide CPU idle.
    :rtype: :py:class:`list`

    Each CPU will be added to a stack group equal to that CPU's number.  CPU idle metrics are hidden by default.
    ``metrics`` is not modified by this function.

    """
    stacked_metrics = deepcopy(metrics)
    for m in stacked_metrics:
        match = CPU_NUMBER_RE.match(m["name"])
        m["stack"] = int(match.group("number"))
        if hide_idle and m["name"].endswith("idle"):
            m["hidden"] = True
    return stacked_metrics


def get_cpu_graph_data(check_bundle):
    """Get graph data for ``check_bundle``.

    :param dict check_bundle: The check bundle to create graph data with.
    :rtype: :py:class:`dict`

    The returned data :py:class:`dict` can be used to :meth:`~circonus.CirconusClient.create` a `graph
    <https://login.circonus.com/resources/api/calls/graph>`_.

    """
    metrics = get_stacked_cpu_metrics(get_cpu_metrics(get_metrics(check_bundle, CPU_METRIC_RE)))
    datapoints = []
    for cid in check_bundle["_checks"]:
        check_id = get_check_id_from_cid(cid)
        datapoints.extend(get_datapoints(check_id, metrics, {"derive": "counter"}))
    custom_data = {"title": "%s cpu" % check_bundle["target"], "max_left_y": 100}
    return get_graph_data(check_bundle, datapoints, custom_data)



"""

circonus.collectd.memory
~~~~~~~~~~~~~~~~~~~~~~~~

Create graph data from a ``collectd`` check bundle.

"""

import re

from circonus.graph import get_graph_data
from circonus.metric import get_datapoints, get_metrics, get_metrics_sorted_by_suffix
from circonus.util import get_check_id_from_cid


MEMORY_METRIC_SUFFIXES = ["used", "buffered", "cached", "free"]
"""Ordered list of metric suffixes.

This list is used to filter and sort metrics in preparation for creating a graph.

"""

MEMORY_METRIC_RE = re.compile(r"""
^memory                         # Starts with "memory"
`.*`                            # Anything in between
({}|{}|{}|{})$                  # Ends with defined suffix
""".format(*MEMORY_METRIC_SUFFIXES), re.X)
"""A compiled regular expression which matches ``collectd`` metrics."""


def get_sorted_memory_metrics(metrics):
    """Get a sorted list of metrics from ``metrics``.

    :param list metrics: The metrics to sort.

    The metrics are sorted by explicit suffix, i.e., :const:`~circonus.collectd.memory.MEMORY_METRIC_SUFFIXES`.

    """
    return get_metrics_sorted_by_suffix(metrics, MEMORY_METRIC_SUFFIXES)


def get_memory_datapoints(check_bundle, metrics):
    """Get a list of datapoints from *sorted* ``metrics``.

    :param dict check_bundle: The check bundle.
    :param list metrics: The sorted metrics to cerate datapoints with.
    :rtype: :py:class:`list`

    """
    datapoints = []
    for i, cid in enumerate(check_bundle["_checks"]):
        check_id = get_check_id_from_cid(cid)
        datapoints.extend(get_datapoints(check_id, metrics, {"derive": "gauge", "stack": i}))
    return datapoints


def get_memory_graph_data(check_bundle):
    """Get graph data for ``check_bundle``.

    :param dict check_bundle: The check bundle to create graph data with.
    :rtype: :py:class:`dict`

    The returned data :py:class:`dict` can be used to :meth:`~circonus.CirconusClient.create` a `graph
    <https://login.circonus.com/resources/api/calls/graph>`_.

    """
    data = {}
    memory_metrics = get_metrics(check_bundle, MEMORY_METRIC_RE)
    if memory_metrics:
        sorted_memory_metrics = get_sorted_memory_metrics(memory_metrics)
        datapoints = get_memory_datapoints(check_bundle, sorted_memory_metrics)
        custom_data = {"title": "%s memory" % check_bundle["target"], "min_left_y": 0, "min_right_y": 0}
        data = get_graph_data(check_bundle, datapoints, custom_data)
    return data



"""

circonus.collectd.interface
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Create graph data from a ``collectd`` check bundle containing
`interface <https://collectd.org/wiki/index.php/Plugin:Interface>`_ metrics.

Transmitted metrics will be on the top of the graph and received metrics will be on the bottom.

"""

from circonus.graph import get_graph_data
from circonus.metric import get_datapoints
from circonus.util import get_check_id_from_cid


DATA_FORMULA_TRANSMITTER = "=8*VAL"
"""The data formula to apply to transmitted interface metrics."""

DATA_FORMULA_RECEIVER = "=-8*VAL"
"""The data formula to apply to received interface metrics."""


def get_interface_metrics(metrics, interface_name="eth0", kind=None):
    """Get metrics for interface ``interface_name`` and ``kind``.

    :param list metrics: The metrics.
    :param str interface_name: (optional) The interface name, e.g., "eth0".
    :param str kind: (optional) The kind of interface metrics to get, e.g., "octets".
    :rtype: :py:class:`list`

    """
    if kind:
        interface_metrics = [m for m in metrics if m["name"].startswith("interface`%s" % interface_name) and
                             kind in m["name"]]
    else:
        interface_metrics = [m for m in metrics if m["name"].startswith("interface`%s" % interface_name)]
    interface_metrics.sort(reverse=True)
    return interface_metrics


def is_transmitter(metric):
    """Is interface ``metric`` a transmitter?

    :param dict metric: The metric.
    :rtype: :py:class:`bool`

    """
    return metric.get("name", "").endswith("tx")


def is_receiver(metric):
    """Is interface ``metric`` a receiver?

    :param dict metric: The metric.
    :rtype: :py:class:`bool`

    """
    return metric.get("name", "").endswith("rx")


def get_interface_datapoints(check_bundle, interface_name="eth0"):
    """Get a list of datapoints for ``check_bundle`` and ``interface_name``.

    :param list check_bundle: The check bundle.
    :param str interface_name: (optional) The interface name, e.g., "eth0".
    :rtype: :py:class:`list`

    ``octets`` and ``errors`` will be returned.  ``octets`` datapoints have data formulas added to them which makes
    them render as bits per second.  Transmitted ``octets`` will be on the top and received ``octets`` will be on the
    bottom of the graph due to the data formulas.

    """
    datapoints = []
    for cid in check_bundle.get("_checks", []):
        check_id = get_check_id_from_cid(cid)
        metrics = check_bundle["metrics"]
        octets = get_interface_metrics(metrics, interface_name, "octets")
        for m in octets:
            if is_transmitter(m):
                m["data_formula"] = DATA_FORMULA_TRANSMITTER
            elif is_receiver(m):
                m["data_formula"] = DATA_FORMULA_RECEIVER
        datapoints.extend(get_datapoints(check_id, octets, {"derive": "counter"}))
        errors = get_interface_metrics(metrics, interface_name, "errors")
        datapoints.extend(get_datapoints(check_id, errors, {"derive": "counter", "axis": "r"}))
    return datapoints


def get_interface_graph_data(check_bundle, interface_name="eth0"):
    """Get graph data for ``check_bundle``.

    :param dict check_bundle: The check bundle to create graph data with.
    :param str interface_name: (optional) The interface name, e.g., "eth0".

    :rtype: :py:class:`dict`

    The returned data :py:class:`dict` can be used to :meth:`~circonus.CirconusClient.create` a `graph
    <https://login.circonus.com/resources/api/calls/graph>`_.

    """
    datapoints = get_interface_datapoints(check_bundle, interface_name)
    custom_data = {"title": "%s interface %s bit/s" % (check_bundle["target"], interface_name)}
    return get_graph_data(check_bundle, datapoints, custom_data)



#!/usr/bin/env python

from setuptools import find_packages, setup

import codecs

import circonus


with codecs.open("README.rst", "r", "utf-8") as f:
    readme = f.read()
with codecs.open("HISTORY.rst", "r", "utf-8") as f:
    history = f.read()

setup(
    name="circonus",
    version=circonus.__version__,
    description="Interact with the Circonus REST API.",
    long_description=readme + "\n\n" + history,
    author="Monetate Inc.",
    author_email="kmolendyke@monetate.com",
    url="https://github.com/monetate/circonus",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7"
    ],
    keywords="circonus monitoring analytics",
    packages=find_packages(),
    install_requires=["colour", "requests"]
)



"""

cicronus.util
~~~~~~~~~~~~~

Utility functions that other modules depend upon.

"""


from posixpath import sep as pathsep

from colour import Color


def get_check_id_from_cid(cid):
    """Get a check id integer from ``cid``.

    :param str cid: The check id.
    :rtype: :py:class:`int`

    """
    return int(cid.strip(pathsep).rpartition(pathsep)[-1])


def get_resource_from_cid(cid):
    """Get the resource name from ``cid``.

    :param str cid: The check id.
    :rtype: :py:class:`str`

    """
    return cid.strip(pathsep).split(pathsep)[0]


def colors(items):
    """Create a generator which returns colors for each item in ``items``.

    :param list items: The list to generate colors for.
    :rtype: generator(`colour.Color <https://pypi.python.org/pypi/colour>`_)

    """
    if len(items) < 2:
        c = (c for c in (Color("red"),))
    else:
        color_from = Color("red")
        color_to = Color("green")
        c = (color_from.range_to(color_to, len(items)))
    return c
